class Router:
    """A class for generating Django-style path definitions based on HTTP methods.

    Methods:
    - generate(http_method, path, controller):
        Generates a Django path definition string based on the provided HTTP method, path, and controller.

    - get(path, controller):
        Generates a path definition for the 'GET' method.

    - post(path, controller):
        Generates a path definition for the 'POST' method.

    - put(path, controller):
        Generates a path definition for the 'PUT' method.

    - delete(path, controller):
        Generates a path definition for the 'DELETE' method.

    - patch(path, controller):
        Generates a path definition for the 'PATCH' method.
    """

    @classmethod
    def generate(cls, http_method, path, controller, show_lists = False):
        """Generate a Django path definition string.

        Args:
        - http_method (str): The HTTP method for the route ('get', 'post', 'put', 'delete', 'patch').
        - path (str): The URL path for the route.
        - controller (callable): The controller function or method.

        Returns:
        str: A Django path definition string.
        """
        class_name = ((controller.__qualname__).split('.'))[0]
        method_name = controller.__name__
        module_name = ((controller.__module__).split('.'))[1]
        path_split = path.split('/')
        path_name = '--'.join(path_split[:-1]) + path_split[-1].rstrip('/')

        route_lists = {
            'method': http_method,
            'path': f'{module_name}/{path}',
            'view': f'{class_name}.{method_name}',
            'name': f'{module_name}_{path_name}',
            'module': module_name
            }
        the_path = f"path('{path}', {class_name}.as_view({{'{http_method}': '{method_name}'}}), name='{module_name}_{path_name}')"
        data =  {
            'route_lists': route_lists,
            'path': the_path
        }
        return  data

    @classmethod
    def get(cls, path, controller):
        """Generate a path definition for the 'GET' method.

        Args:
        - path (str): The URL path for the route.
        - controller (callable): The controller function or method.

        Returns:
        str: A Django path definition string for the 'GET' method.
        """
        return cls.generate('get', path, controller)

    @classmethod
    def post(cls, path, controller):
        """Generate a path definition for the 'POST' method.

        Args:
        - path (str): The URL path for the route.
        - controller (callable): The controller function or method.

        Returns:
        str: A Django path definition string for the 'POST' method.
        """
        return cls.generate('post', path, controller)

    @classmethod
    def put(cls, path, controller):
        """Generate a path definition for the 'PUT' method.

        Args:
        - path (str): The URL path for the route.
        - controller (callable): The controller function or method.

        Returns:
        str: A Django path definition string for the 'PUT' method.
        """
        return cls.generate('put', path, controller)

    @classmethod
    def delete(cls, path, controller):
        """Generate a path definition for the 'DELETE' method.

        Args:
        - path (str): The URL path for the route.
        - controller (callable): The controller function or method.

        Returns:
        str: A Django path definition string for the 'DELETE' method.
        """
        return cls.generate('delete', path, controller)

    @classmethod
    def patch(cls, path, controller):
        """Generate a path definition for the 'PATCH' method.

        Args:
        - path (str): The URL path for the route.
        - controller (callable): The controller function or method.

        Returns:
        str: A Django path definition string for the 'PATCH' method.
        """
        return cls.generate('patch', path, controller)


class Route:
    """A class representing a collection of routes.

    Args:
    - routes (list): A list of routes.

    Methods:
    - get_all_routes():
        Retrieves all routes in the collection, combining routes with the same path.

    Returns:
    list: A list of combined route paths.
    """

    def __init__(self, routes):
        self.routes = routes

    def get_all_routes(self):
        """Retrieve all routes in the collection, combining routes with the same path.

        Returns:
        list: A list of combined route paths.
        """
        import random

        # Group routes by path
        path_groups = {}
        for route in self.routes:
            path = route["path"].split("path('")[1].split("'")[0]
            view_info = route["path"].split("as_view({")[1].split("})")[0]
            method_info = view_info.strip("'{}")
            method, func = method_info.split("': '")
            
            if path in path_groups:
                # Add method to existing path
                path_groups[path]["methods"][method] = func
            else:
                # Create new path entry
                class_name = route["path"].split("path('")[1].split(".as_view")[0].split(", ")[-1]
                module_name = route["path"].split("name='")[1].split("_")[0]
                
                path_segments = path.split("/")
                path_segments = [seg.lower() for seg in path_segments if seg and '<' not in seg]
                segments_part = '_'.join(path_segments)
                random_number = str(random.randint(1000000000, 9999999999))
                
                path_groups[path] = {
                    "class_name": class_name,
                    "methods": {method: func},
                    "module_name": module_name,
                    "path_segments": path_segments,
                    "random_number": random_number
                }
        
        # Generate combined paths
        combined_routes = []
        for path, info in path_groups.items():
            methods_dict = str(info["methods"]).replace("'", '"')
            # Sort methods alphabetically and join with underscore
            methods_str = '_'.join(sorted(info["methods"].keys()))
            segments_part = '_'.join(info["path_segments"])
            name = f"{info['module_name']}__{segments_part}__{methods_str}_{info['random_number']}"
            
            combined_path = f"path('{path}', {info['class_name']}.as_view({methods_dict}), name='{name}')"
            combined_routes.append(combined_path)
            
        return combined_routes

    def show_lists(self):
        return [route["route_lists"] for route in self.routes]


import uuid
from django.db import models
from .managers import SoftDeleteManager


class TimeStampedModel(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class Extensions(models.Model):
    """ Best practice for lookup field url instead pk or slug """

    uuid = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class UUIDModel(models.Model):
    """ 
    Best practice for lookup field url instead pk or slug.
    for security
    """

    uuid = models.UUIDField(db_index=True, default=uuid.uuid4, editable=False)

    class Meta:
        abstract = True


class SoftDeleteModel(models.Model):
    is_deleted = models.BooleanField(default=False)

    objects = SoftDeleteManager()

    class Meta:
        abstract = True


from rest_framework.response import Response
from rest_framework.pagination import PageNumberPagination
from collections import OrderedDict

class LaravelStylePagination(PageNumberPagination):
    page_size = 10
    
    def get_paginated_response(self, data, wrap):
        data =  (OrderedDict([
            ('per_page', self.page_size),
            ('current_page', self.page.number),
            ('last_page', self.page.paginator.num_pages),
            ('next_page_url', self.get_next_link() or None),
            ('prev_page_url', self.get_previous_link() or None),
            ('total', self.page.paginator.count),
            ('from', self.page.start_index()),
            ('to', self.page.end_index()),
            (wrap, data),
        ]))
        return data

        
def paginate(request, queryset, serializer_class, per_page=2, wrap='data'):
    # Create an instance of the custom pagination class
    paginator = LaravelStylePagination()
    paginator.page_size = per_page

    # Paginate the queryset
    page = paginator.paginate_queryset(queryset, request)

    # Serialize the paginated data using the provided serializer class
    serializer = serializer_class(page, many=True)

    # Return the paginated data along with pagination metadata
    return paginator.get_paginated_response(serializer.data, wrap)



from collections import OrderedDict
from math import ceil
from django.core.paginator import Paginator
from rest_framework.response import Response
from urllib.parse import urlparse

def paginate(request, queryset, serializer_class, per_page=10, wrap='data', additional_data=None):
    # Get per_page from query params if present, otherwise use default
    per_page = int(request.query_params.get('per_page', per_page))
    page = int(request.query_params.get('page', 1))
    total = queryset.count()
    last_page = ceil(total / per_page)
    offset = (page - 1) * per_page

    paginated_queryset = queryset[offset:offset + per_page]
    serializer = serializer_class(paginated_queryset, many=True)
    data = serializer.data

    full_path = str(request.build_absolute_uri())
    parsed_url = urlparse(full_path)
    path_without_query = f"{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}"

    first_page_url = f"{path_without_query}?page=1&per_page={per_page}"
    last_page_url = f"{path_without_query}?page={last_page}&per_page={per_page}"
    next_page_url = f"{path_without_query}?page={page + 1}&per_page={per_page}" if page < last_page else None
    prev_page_url = f"{path_without_query}?page={page - 1}&per_page={per_page}" if page > 1 else None

    # Convert all URLs to strings explicitly
    response_data = OrderedDict([
        (wrap, data),
        ('total', total),
        ('per_page', int(per_page)),  # ensure integer
        ('current_page', int(page)),   # ensure integer
        ('last_page', int(last_page)), # ensure integer
        ('first_page_url', str(first_page_url)),
        ('last_page_url', str(last_page_url)),
        ('next_page_url', str(next_page_url) if next_page_url else None),
        ('prev_page_url', str(prev_page_url) if prev_page_url else None),
    ])

    # Merge additional_data if provided
    if additional_data:
        response_data['additional_data'] = additional_data

    return response_data


import sys
import platform
import socket
import os

def system_environment():
        python_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        default_encoding = sys.getdefaultencoding()
        file_system_encoding = sys.getfilesystemencoding()
        os_name = platform.system()
        if os_name == "Darwin":
            os_name = "macOS"

        # when VPN is enabled it can fail for some VPN clients on macOS
        try:
            ip = socket.gethostbyname(socket.gethostname())
        except socket.gaierror:
            print(
                "Exceptionite did not manage to fetch the IP address. Disable your VPN or add "
                + "'127.0.0.1 YOUR_HOSTNAME' line in /etc/hosts file."
            )
            ip = "Error fetching the IP address (open your terminal)"

        return {
            "Python Version": python_version,
            "Python Interpreter": sys.executable,
            "Virtual env": os.getenv("VIRTUAL_ENV"),
            "Python argv": sys.argv,
            "Working Dir": os.getcwd(),
            "OS": os_name,
            "Arch": platform.architecture()[0],
            "Host Name": socket.gethostname(),
            "IP": ip,
            "File System Encoding": file_system_encoding,
            "Default Encoding": default_encoding,
        }


from django.db import models

# class APILogs(models.Model):
#     timestamp = models.DateTimeField(auto_now_add=True)
#     url = models.URLField()
#     method = models.CharField(max_length=10)
#     request_headers = models.TextField()
#     request_body = models.TextField()
#     status_code = models.PositiveIntegerField()
#     response_content = models.TextField()
#     api_call_time = models.FloatField()
#     server_execution_time = models.FloatField()
#     client_ip = models.GenericIPAddressField()

#     class Meta:
#         db_table = 'django_petra_api_logs'



from django.utils import timezone
from django_petra.project.models import APILogs

class APILoggerMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        start_time = timezone.now()
        response = self.get_response(request)
        end_time = timezone.now()
        api_call_time = (end_time - start_time).total_seconds()

        # Log API request and response to the database
        if 'hello-django-petra/api/show-api-logs/' not in request.path :
            self.log_api(request, response, api_call_time)

        return response

    def log_api(self, request, response, api_call_time):

        log_entry = APILogs(
            url=request.build_absolute_uri(),
            method=request.method,
            request_headers=str(request.headers),
            request_body=str(request.body),
            status_code=response.status_code,
            response_content=str(response.content),
            api_call_time=api_call_time,
            server_execution_time=0.0,  # Placeholder for server execution time
            client_ip=request.META.get('REMOTE_ADDR', ''),
        )
        log_entry.save()
        print("yes")



from rest_framework.exceptions import ParseError

def get_request_data(request, key, default_value=None):
    """
    Extracts data from request.data, request.FILES, and request.query_params based on the request method and content type.

    :param request: Petra request object
    :param key: The key to look for in the request data
    :param default_value: The default value to return if the key is not found
    :return: The value associated with the key, or the default value if not found
    """
    try:
        if request.method in ['POST', 'PUT', 'PATCH']:
            if request.content_type.startswith('multipart/form-data'):
                # For file uploads
                if key in request.FILES:
                    return request.FILES.get(key)
                elif key in request.data:
                    return request.data.get(key)
            elif request.content_type in ['application/json', 'application/x-www-form-urlencoded']:
                # For JSON or form-encoded data
                if key in request.data:
                    return request.data.get(key)
        elif request.method == 'GET':
            if key in request.query_params:
                return request.query_params.get(key)
        else:
            if key in request.query_params:
                return request.query_params.get(key)
              
        return default_value
    except Exception as e:
        raise ParseError(f"Error extracting request data for key '{key}': {str(e)}")

def get_all_request_data(request):
    """
    Combines all request data from query_params, request.data, and FILES into a single dictionary.
    
    :param request: Petra request object
    :return: Dictionary containing all request data
    """
    try:
        combined_data = {}
        
        # Add query parameters
        combined_data.update(request.query_params.dict())
        
        # Add request data for POST/PUT/PATCH methods
        if request.method in ['POST', 'PUT', 'PATCH']:
            if request.content_type.startswith('multipart/form-data'):
                # Add both form data and files
                combined_data.update(request.data.dict())
                combined_data.update(request.FILES.dict())
            elif request.content_type in ['application/json', 'application/x-www-form-urlencoded']:
                # Add JSON or form-encoded data
                combined_data.update(request.data)
                
        return combined_data
    except Exception as e:
        raise ParseError(f"Error extracting all request data: {str(e)}")

def get_request_headers(request):
    """
    Gets all headers from the request object.
    
    :param request: Petra request object
    :return: Dictionary containing all request headers
    """
    try:
        # META keys starting with 'HTTP_' contain the HTTP headers
        headers = {
            key[5:].lower().replace('_', '-'): value 
            for key, value in request.META.items() 
            if key.startswith('HTTP_')
        }
        
        # Special handling for Content-Type and Content-Length as they don't have HTTP_ prefix
        if 'CONTENT_TYPE' in request.META:
            headers['content-type'] = request.META['CONTENT_TYPE']
        if 'CONTENT_LENGTH' in request.META:
            headers['content-length'] = request.META['CONTENT_LENGTH']
            
        return headers
    except Exception as e:
        raise ParseError(f"Error extracting request headers: {str(e)}")



import os

class DatabaseConfig:
    @classmethod
    def get_database_settings(cls):
        db_engine = os.getenv('DB_CONNECTION', '')
        db_name = os.getenv('DB_DATABASE', '')
        db_user = os.getenv('DB_USERNAME', '')
        db_password = os.getenv('DB_PASSWORD', '')
        db_host = os.getenv('DB_HOST', '')
        db_port = os.getenv('DB_PORT', '')

        return {
            'ENGINE': cls.get_database_engine(db_engine),
            'NAME': db_name,
            'USER': db_user,
            'PASSWORD': db_password,
            'HOST': db_host,
            'PORT': db_port,
        }

    @staticmethod
    def get_database_engine(db_engine):
        if db_engine == 'mysql':
            return 'django.db.backends.mysql'
        elif db_engine == 'postgres':
            return 'django.db.backends.postgresql'
        else:
            raise ValueError(f"Unsupported database engine: {db_engine}")




import os
from django.conf import settings

def base_path(*args):
    """Return the absolute path to the Django project's base directory."""
    return os.path.join(settings.BASE_DIR, *args)
  
def modules_path(*args):
  return base_path('modules', *args)

def mail_template_path(*args):
  return base_path('template', 'mail', *args)

def base_module_name():
    root_urlconf = settings.ROOT_URLCONF
    base_module_name = root_urlconf.split('.')[0]
    return base_module_name


from rest_framework.viewsets import ViewSet as View
from rest_framework.response import Response
from rest_framework.serializers import ModelSerializer
from django_petra.status import Status as MyStatus
from django_petra.petra_core.permissions import Permissions

class ViewSet(View, Permissions):
    def __init__(self, *args, **kwargs):
        View.__init__(self, *args, **kwargs)
        Permissions.__init__(self)

class Response(Response):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
class status(MyStatus):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
class ModelSerializer(ModelSerializer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


import os
import django
import importlib
from django.conf import settings
from django_petra.core import bind_modules_app, bind_modules_urls
from django_petra.path import base_module_name

def init_cors_middleware():
    cors_middleware_class = 'django_petra.cors.middleware.CorsMiddleware'
    settings.MIDDLEWARE.insert(0, cors_middleware_class)
    
def init_all_modules():
  settings.INSTALLED_APPS.extend(bind_modules_app())
  
def init_module_urls():
  the_django_setup()
  the_base_module_name = base_module_name()
  the_urls = importlib.import_module(f"{the_base_module_name}.urls")
  urlpatterns = the_urls.urlpatterns
  urlpatterns.extend(bind_modules_urls())
  
def the_django_setup():
  the_base_module_name = base_module_name()
  os.environ.setdefault('DJANGO_SETTINGS_MODULE', f'{the_base_module_name}.settings')
  django.setup()


from django_petra.petra_core import Response, ViewSet, status
from django_petra.core import show_modules_url
from django_petra.raw_query.helpers import (
  fetch_all_to_dictionary,
  raw_query_collection
)
from django.db import connection

class ProjectViewset(ViewSet):
  
  def get_status(self, request):
    data = {
      'status': "django-petra is working fine"
    }
    return Response(data, status=status.HTTP_200_OK)
  
  def get_routes(self, request):
    data = {
      'routes': show_modules_url()
    }
    
    return Response(data, status=status.HTTP_200_OK)
  
  def get_api_logs(self, request):
      query = 'SELECT * FROM django_petra_api_logs'
  
      with connection.cursor() as cursor:
          cursor.execute(query)
          results = fetch_all_to_dictionary(cursor)
          output = raw_query_collection(request, results, 'api_logs')
      return Response(output)
  
  def get_installed_packages(self, request):
    import pkg_resources
    packages = pkg_resources.working_set
    package_info = []

    for package in packages:
      package_name = package.key
      installed_version = package.version
  
      package_info.append({
        'project_name': package.project_name,
        'name': package_name,
        'version': installed_version,
        'pypi_url': f'https://pypi.org/project/{package_name}/{installed_version}',
        
      })
    
    return Response(package_info, status=status.HTTP_200_OK)
  
  def get_system_environment(self, request):
    from django_petra.project.utils.system_environment import system_environment
    system_info = system_environment()
    return Response(system_info, status=status.HTTP_200_OK)
        


from django_petra.project.tasks import petra_mail

class SendMailQueue:
    def __init__(self):
        self.email = {
            "delay": 10
        }

    def from_email(self, sender):
        self.email["from"] = sender
        return self

    def to(self, recipients):
        self.email["to"] = recipients if isinstance(recipients, list) else [recipients]
        return self

    def cc(self, cc_recipients):
        self.email["cc"] = cc_recipients if isinstance(cc_recipients, list) else [cc_recipients]
        return self

    def bcc(self, bcc_recipients):
        self.email["bcc"] = bcc_recipients if isinstance(bcc_recipients, list) else [bcc_recipients]
        return self

    def subject(self, subject):
        self.email["subject"] = subject
        return self

    def template(self, template_name, context):
        self.email["template_name"] = template_name
        self.email["context"] = context
        return self

    def attach(self, file_paths):
        self.email.attach_files = file_paths
        return self
    
    def delay(self, the_delay = 2):
        self.email["delay"] = the_delay
        return self

    def send(self):
        """
        Send the email using the configured EmailMessage instance.

        Returns:
            None
        """
        petra_mail.apply_async(args=[self.email], countdown=self.email["delay"])



from django.urls import path 
from django_petra.project.views import ProjectViewset
from django_petra.router import Router, Route

routes = [
    Router.get('status/', ProjectViewset.get_status),
    Router.get('api/show-routes/', ProjectViewset.get_routes),
    Router.get('api/show-api-logs/', ProjectViewset.get_api_logs),
    Router.get('api/get-installed-packages', ProjectViewset.get_installed_packages),
    Router.get('api/get-system-environment', ProjectViewset.get_system_environment)
]

urlpatterns = [eval(route) for route in Route(routes).get_all_routes()]


